#!/usr/bin/php
<?php
/** DomPHPTests
  * @package domphptests
  * @author Dominique Fournier <dominique@fournier38.fr>
  */

class PHPUnit_Framework_TestCase extends DomPHPTests
{
}

/** Allow to do unit tests on PHP softs */
class DomPHPTests
{
  /** Set the debug to true to display the result of the test
    */
  public $debug = false;
  /** Set the time to true to get the time need to pass all the tests
    */
  public $time = false;
  /** The result of the output if tested
    */
  private $expect;
  /** The property use to store the error messages
    */
  private $error;
  /** The number of tests executed (passed or failed)
    */
  private $countTest = 0;
  /** The number of tests passwed with error
    */
  private $countError = 0;
  /** Flag when a test is done (with or without success)
    */
  private $testDone;
  /** The CoverStats are requested
    */
  public $coverStats;
  /** The Cover path to store the coverage HTML reports
    */
  public $coverPath;
  /** The basedir of the project
    */
  public $baseDir;
  /** The bootstrap use with the test. To be removed from the coverage
    */
  public $boostrap;

  /** If there is an error raised, catch it and display it in debug
    * @param integer $errno The error number
    * @param string $errstr The error message
    * @param string $errfile The file with the error
    * @param integer $errline The line with the error in file
    */
  public function catchErrors ($errno, $errstr, $errfile, $errline)
  {
    $msg = "";
    if (!(error_reporting() & $errno))
    {
      // This error code is not included in error_reporting
      return;
    }
    switch ($errno)
    {
    case E_NOTICE:            $msg .= "PHP Notice: ";break;
    case E_WARNING:           $msg .= "PHP Warning: ";break;
    case E_ERROR:             $msg .= "PHP Error: ";break;
    case E_PARSE:             $msg .= "PHP Parse Error: ";break;
    case E_CORE_ERROR:        $msg .= "PHP Core Error: ";break;
    case E_CORE_WARNING:      $msg .= "PHP Core Warning: ";break;
    case E_COMPILE_ERROR:     $msg .= "PHP Compile Error: ";break;
    case E_COMPILE_WARNING:   $msg .= "PHP Compile Warning: ";break;
    case E_STRICT:            $msg .= "PHP Strict: ";break;
    case E_RECOVERABLE_ERROR: $msg .= "PHP Recoverable Error: ";break;
    case E_DEPRECATED:        $msg .= "PHP Deprecated Error: ";break;
    // Do not break if the user generate an error
    case E_USER_ERROR:        return true;
    case E_USER_WARNING:      return true;
    case E_USER_NOTICE:       return true;
    default:
      echo "ERROR domphptests : missing errno $errno\n";exit;
    }
    $msg .= "$errstr in $errfile on line $errline\n";
    if ($this->debug === false)
      throw new \Exception ($msg);
    else
      echo $msg;
    return true;
  }

  /** Check if $condition is false. Then if not, return the message $message
    * @param mixed $condition The condition to check
    * @param string $message The message to return if the $val1 !== false
    */
  public function assertFalse ($condition, $message = "%s === false")
  {
    $this->testDone = true;
    if ($condition !== false)
      $this->error = sprintf ($message, var_export ($condition, true));
  }

  /** Check if $val1 > $val2. Then if not, return the message $message
    * @param mixed $val1 The val1 to check
    * @param mixed $val2 The val2 to check
    * @param string $message The message to return if the $val1 <= $val2
    */
  public function assertGreaterThan ($val1, $val2, $message = "%s > %s")
  {
    $this->testDone = true;
    if ($val1 <= $val2)
      $this->error = sprintf ($message, var_export ($val1, true),
                              var_export ($val2, true));
  }

  /** Check if $val1 >= $val2. Then if not, return the message $message
    * @param mixed $val1 The val1 to check
    * @param mixed $val2 The val2 to check
    * @param string $message The message to return if the $val1 < $val2
    */
  public function assertGreaterThanOrEqual ($val1, $val2, $message = "%s >= %s")
  {
    $this->testDone = true;
    if ($val1 < $val2)
      $this->error = sprintf ($message, var_export ($val1, true),
                              var_export ($val2, true));
  }

  /** Check if $val1 < $val2. Then if not, return the message $message
    * @param mixed $val1 The val1 to check
    * @param mixed $val2 The val2 to check
    * @param string $message The message to return if the $val1 >= $val2
    */
  public function assertLessThan ($val1, $val2, $message = "%s < %s")
  {
    $this->testDone = true;
    if ($val1 >= $val2)
      $this->error = sprintf ($message, var_export ($val1, true),
                              var_export ($val2, true));
  }

  /** Check if $val1 <= $val2. Then if not, return the message $message
    * @param mixed $val1 The val1 to check
    * @param mixed $val2 The val2 to check
    * @param string $message The message to return if the $val1 > $val2
    */
  public function assertLessThanOrEqual ($val1, $val2, $message = "%s <= %s")
  {
    $this->testDone = true;
    if ($val1 > $val2)
      $this->error = sprintf ($message, var_export ($val1, true),
                              var_export ($val2, true));
  }

  /** Check if $val1 ~= $val2. Then if not, return the message $message
    * @param mixed $regex The regex to apply
    * @param mixed $val The result of the test to check
    * @param string $message The message to return if the $val1 !~= $val2
    */
  public function assertRegexp ($regex, $val, $message = "%s ~= %s")
  {
    $this->testDone = true;
    if (preg_match ($regex, $val) !== 1)
      $this->error = sprintf ($message, var_export ($val, true),
                              var_export ($regex, true));
  }

  /** Check if $val1 === $val2. Then if not, return the message $message
    * @param mixed $val1 The result to check
    * @param mixed $val2 The second parameter
    * @param string $message The message to return if the $val1 !== $val2
    */
  public function assertSame ($val1, $val2, $message = "%s === %s")
  {
    $this->testDone = true;
    if ($val1 !== $val2)
      $this->error = sprintf ($message, var_export ($val1, true),
                              var_export ($val2, true));
  }

  /** Check if $condition is true, then if not, return the message $message
    * @param mixed $condition The result to check
    * @param string $message The message to return if the condition is not true
    */
  public function assertTrue ($condition, $message = "%s === true")
  {
    $this->testDone = true;
    if ($condition !== true)
      $this->error = sprintf ($message, var_export ($condition, true));
  }

  /** Check if an output is generated by the code and is match the
    * parameter regexp
    * @param string $expectedRegex The regex that match the output of the
    *                              test
    */
  public function expectOutputRegex ($expectedRegex)
  {
    $this->testDone = true;
    $this->expect = array ("expectOutputRegex", $expectedRegex);
  }

  /** Check if an output is generated by the code and is the same as the
    * parameter
    * @param string $expectedString The output that should be returned by the
    *                               test
    */
  public function expectOutputString ($expectedString)
  {
    $this->testDone = true;
    $this->expect = array ("expectOutputString", $expectedString);
  }

  /** Check if an exception is well raised by the test
    * @param string|null $class The exception class name to be raised
    * @param string|null $message The message raised in the Exception
    * @param integer|null $code The error code raised in the Exception
    */
  public function expectException ($class = null, $message = null,
                                   $code = null)
  {
    $this->testDone = true;
    $this->expect = array ("expectException", $class, $message, $code);
  }

  /** Check if an exception is well raised by the test
    * @param string|null $class The exception class name to be raised
    * @param string|null $message The message raised in the Exception
    * @param integer|null $code The error code raised in the Exception
    */
  public function setExpectedException ($class = null, $message = null,
                                        $code = null)
  {
    $this->testDone = true;
    $this->expectException ($class, $message, $code);
  }

  /** The main method launch the tests for a test file
    * @param string $fileTest The test file to run
    * @param string $errorlogFile The error file to fill
    * @param string $statFile The stat error file to fill
    */
  public function main ($fileTest, $errorlogFile, $statFile)
  {
    // TODO : Check if there is a PHP syntax error
    if ($this->debug) echo "Start of test $fileTest\n";
    set_error_handler (array ($this, "catchErrors"));
    $startTime = microtime (true);
    $errorLog = "";
    require_once ($fileTest);
    foreach (get_declared_classes () as $class)
    {
      if (!is_subclass_of ($class, "DomPHPTests"))
        continue;
      // The reflection test is needed for sequential tests : appending all the
      // files in the memory without clearing it
      $reflectionClass = new ReflectionClass ($class);
      if (substr ($reflectionClass->getFileName (), 1+strlen (getcwd())) !==
          $fileTest)
        continue;
      unset ($obj);
      $obj = new $class ();
      foreach (get_class_methods ($class) as $method)
      {
        // If the method don't start by 'test', don't test it !
        if (substr ($method, 0, 4) !== "test")
          continue;
        $this->countTest++;
        if ($this->debug && $this->time) echo date ("H:i:s")." ";
        if ($this->debug) echo "Test in $fileTest class $class::$method : ";
        unset ($e);
        unset ($exceptionclass);
        unset ($exceptionmessage);
        unset ($exceptioncode);
        unset ($obj->error);
        unset ($obj->expect);
        unset ($obj->testDone);
        try
        {
          ob_start ();
          unset ($coverTmp);
          if (function_exists ("xdebug_start_code_coverage"))
            xdebug_start_code_coverage ();
          $obj->$method ();
          if (function_exists ("xdebug_get_code_coverage"))
            $coverTmp = xdebug_get_code_coverage ();
          if (function_exists ("xdebug_stop_code_coverage"))
            xdebug_stop_code_coverage (false);
          $output = ob_get_contents ();
          ob_end_clean ();
        }
        catch (Exception $e)
        {
          $output = ob_get_contents ();
          ob_end_clean ();
          $exceptionclass = get_class ($e);
          $exceptionmessage = $e->getMessage ();
          $exceptioncode = $e->getCode ();
        }

        if ($this->debug && strlen ($output)) echo "\n$output";
        // Manage the expect tests (tests executed after the complete execution)
        if (isset ($obj->expect[0]))
        {
          switch ($obj->expect[0])
          {
          case "expectException":
            $this->testDone = true;
            $expectClass = $obj->expect[1];
            $expectMessage = $obj->expect[2];
            $expectCode = $obj->expect[3];
            if (! isset ($e))
            {
              $obj->error = "Exception is raised";
            }
            elseif ($expectClass !== null && $expectClass !== $exceptionclass)
            {
              $obj->error = "exception class '$exceptionclass' ".
                             "=== '$expectClass'";
            }
            elseif ($expectMessage !== null &&
                    $expectMessage !== $exceptionmessage)
            {
              $obj->error = "exception message ".
                             "'$exceptionmessage' === '$expectMessage'";
            }
            elseif ($expectCode !== null && $expectCode !== $exceptioncode)
            {
              $obj->error = "exception code '$exceptioncode' ".
                             "=== '$expectCode'";
            }
            break;
          case "expectOutputRegex":
            $this->testDone = true;
            $message = "%s === %s";
            $expectedRegex = $obj->expect[1];
            if (preg_match ($expectedRegex, $output) !== 1)
            {
              $obj->error = sprintf ($message,
                                     var_export ($expectedRegex, true),
                                     var_export ($output, true));
            }
            break;
          case "expectOutputString":
            $this->testDone = true;
            $message = "%s === %s";
            $expectedString = $obj->expect[1];
            if ($expectedString !== $output)
            {
              $obj->error = sprintf ($message,
                                     var_export ($expectedString, true),
                                     var_export ($output, true));
            }
            break;
          default:
            die ($obj->expect[0]." NOT FOUND\n");
          }
        }

        if (! isset ($obj->testDone) && isset ($e))
        {
          $this->countError++;
          $errorLog .= "$this->countError) $class::$method\n";
          $errorLog .= "A not waited exception is raised : ".$e->getMessage ().
                       "\n\n";
          if ($this->color) echo "\033[41m\033[1;37m";
          echo "E";
          if ($this->color) echo "\033[0m";
        }
        elseif (isset ($obj->error))
        {
          $this->countError++;
          $errorLog .= "$this->countError) $class::$method\n";
          $errorLog .= "Failed assert that $obj->error\n\n";
          if ($this->color) echo "\033[41m\033[1;37m";
          echo "E";
          if ($this->color) echo "\033[0m";
        }
        elseif (isset ($obj->testDone) && $obj->testDone === true)
        {
          echo ".";
        }
        else
        {
          echo "N";
        }
        if ($this->debug) echo "\n";

        if (isset ($coverTmp))
        {
          // Manage the code coverage: remove the domphptests and the tests
          // themself
          if (key_exists (__FILE__, $coverTmp))
            unset ($coverTmp[__FILE__]);
          if (key_exists (realpath ($this->bootstrap), $coverTmp))
            unset ($coverTmp[realpath ($this->bootstrap)]);
          if (key_exists (realpath ($fileTest), $coverTmp))
            unset ($coverTmp[realpath ($fileTest)]);
          foreach ($coverTmp as $file=>$used)
          {
            if (substr ($file, 0, strlen ($this->baseDir)) !== $this->baseDir ||
              substr ($file, -4) !== ".php")
              unset ($coverTmp[$file]);
          }
          if (! isset ($cover))
            $cover = array ();
          // Merge the local coverage to the main cover. If the old values was
          // -1 (not tested), and the new is 1 (tested), update the $cover
          foreach ($coverTmp as $file=>$values)
          {
            foreach ($values as $line=>$cov)
            {
              if ($cov === 1 && (! key_exists ($file, $cover) ||
                  ! key_exists ($line, $cover[$file]) ||
                  $cover[$file][$line] !== 1))
              {
                $cover[$file][$line] = $cov;
              }
            }
          }
        }
      }
    }
    if ($errorLog !== "")
    {
      file_put_contents ($errorlogFile, "\n");
      if ($this->countError > 0 && $this->color)
        file_put_contents ($errorlogFile, "\033[41m\033[1;37m", FILE_APPEND);
      file_put_contents ($errorlogFile,
                         "$this->countError errors detected in $file on ".
                         "$this->countTest tests", FILE_APPEND);
      if ($this->color)
        file_put_contents ($errorlogFile, "\033[0m", FILE_APPEND);
      file_put_contents ($errorlogFile, "\n", FILE_APPEND);
      file_put_contents ($errorlogFile, $errorLog, FILE_APPEND);
    }
    if ($this->debug) echo "End of test $file : $this->countError errors\n";
    // $coverTests Contains the lines=>array of tests matching the line
    $coverTests = array ();
    $coverage = array ();
    if ($this->coverPath !== false || $this->coverStats !== false)
    {
      foreach ($cover as $file=>$values)
      {
        foreach ($values as $line=>$val)
        {
          $coverTests[$file][$line][] = $fileTest;
        }
      }
      // Generate an array with all the lines of the file in key and
      // all the associated tokens to the value
      foreach ($cover as $file=>$used)
      {
        $content = token_get_all (file_get_contents ($file));
        $tokens = array ();
        $lastLine = 1;
        foreach ($content as $key=>$val)
        {
          if (is_array ($val))
          {
            $nbLines = substr_count ($content[$key][1], "\n");
            for ($i = $content[$key][2] ; $i <= $content[$key][2] + $nbLines ;
                 $i++)
            {
              if (! key_exists ($i, $tokens))
                $tokens[$i] = "";
            }
            for ($i = $content[$key][2] ; $i <= $content[$key][2] + $nbLines ;
                 $i++)
            {
              if (! key_exists ($i, $tokens))
                $tokens[$i] = "";
              if (token_name ($val[0]) !== "T_WHITESPACE")
                $tokens[$i] .= token_name ($val[0])." ";
              $lastLine = $i;
            }
          }
          else
          {
            $nbLines = substr_count ($val, "\n");
            if ($nbLines >= 1) $nbLines --;
            for ($i = $lastLine ; $i <= $lastLine + $nbLines ;
                 $i++)
            {
              if (! key_exists ($i, $tokens))
                $tokens[$i] = "";
              $tokens[$i] .= "$val ";
            }
          }
        }
        // Remove the last line wich contains only a carriage return
        array_pop ($tokens);

        // Parenthesis Management. The parenthesis can be on multiple lines.
        // Force the Coverage of the first open parenthesis to the last closed
        // one.
        // Can be used in function parameters, Exception with long message...
        $parenthesis = 0;
        $parenthesisCover = false;
        // Code coverage: Add the lines which can not be tested with coverage -2
        // The NEW on multilines are tested to color the multilines as the NEW
        // token
        foreach ($tokens as $line=>$val)
        {
          if (substr_count ($val, "("))
          {
            $parenthesis += substr_count ($val, "(");
          }
          if (substr_count ($val, ")"))
          {
            $parenthesis -= substr_count ($val, ")");
          }
          if ($parenthesis > 0 && $parenthesisCover === false)
          {
            // The "T_FUNCTION" is not detected by XDebug, so set the function
            // line to -2;
            if (strpos ($val, "T_FUNCTION ") !== false ||
              // The properties of a class can contains parenthesis
              strpos ($val, "T_PUBLIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PRIVATE T_VARIABLE") === 0 ||
              strpos ($val, "T_PROTECTED T_VARIABLE") === 0 ||
              strpos ($val, "T_PUBLIC T_STATIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PRIVATE T_STATIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PROTECTED T_STATIC T_VARIABLE") === 0)
            {
              $cover[$file][$line] = -2;
            }
            $parenthesisCover = (isset ($cover[$file][$line])) ?
                                 $cover[$file][$line] : -1;
            $cover[$file][$line] = $parenthesisCover;
          }
          elseif ($parenthesis > 0)
          {
            $cover[$file][$line] = $parenthesisCover;
          }
          if ($parenthesis === 0 && $parenthesisCover !== false)
          {
            $cover[$file][$line] = $parenthesisCover;
            $parenthesisCover = false;
          }
          if (key_exists ($line, $cover[$file]))
            continue;
          else
          {
            $cover[$file][$line] = -2;
          }
          if ($val === "T_DOC_COMMENT " ||
              $val === "T_COMMENT " ||
              $val === "T_COMMENT T_COMMENT " ||
              $val === "T_COMMENT T_DOC_COMMENT " ||
              $val === "T_OPEN_TAG " ||
              $val === "T_OPEN_TAG T_DOC_COMMENT " ||
              $val === "T_ELSE " ||
              strpos ($val, "T_SWITCH ") !== false ||
              strpos ($val, "T_TRY ") !== false ||
              // The Class definition
              strpos ($val, "T_CLASS ") !== false ||
              // The Function definition
              strpos ($val, "T_FUNCTION ") !== false ||
              // The properties of a class
              strpos ($val, "T_PUBLIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PRIVATE T_VARIABLE") === 0 ||
              strpos ($val, "T_PROTECTED T_VARIABLE") === 0 ||
              strpos ($val, "T_PUBLIC T_STATIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PRIVATE T_STATIC T_VARIABLE") === 0 ||
              strpos ($val, "T_PROTECTED T_STATIC T_VARIABLE") === 0 ||
              $val === "} " || $val === "{ " ||
              $val === "")
          {
            // These tokens can not be tested by XDebug: keep them in they're
            // old value (probably -2)
            continue;
          }
          $cover[$file][$line] = -1;
        }
        ksort ($cover[$file], SORT_NUMERIC);
      }
      foreach ($cover as $file=>$used)
      {
        if (substr ($file, 0, strlen ($this->baseDir)) === $this->baseDir &&
          substr ($file, -4) === ".php")
          $fileNames[] = substr ($file, strlen ($this->baseDir) + 1);
      }
      $coverage = array (
        "Cover" => $cover,
        "FileNames" => $fileNames,
        "CoverTests" => $coverTests,
      );

    }
    // Statistics
    file_put_contents ($statFile,
                       json_encode (array (
                         "MaxMemory" => memory_get_peak_usage (),
                         "ExecuteTime" => (microtime (true) - $startTime),
                         "TestsNb" => $this->countTest,
                         "Coverage" => $coverage,
                       )));
    return ($this->countError);
  }
}

///////////////////
//   MAIN PART   //
///////////////////
require_once ("domframework/version.php");
if (!defined ("DOMFRAMEWORK_VERSION") ||
    version_compare (DOMFRAMEWORK_VERSION, "0.36") < 0)
  die ("Your DomFramework version is too old : 0.36 minimum\n");

// Change the name of the classes in Domframework to allow the Domframework
// to be tested by domphptests and used in domphptests
eval (str_replace ("verify", "DPT_verify",
      strstr (file_get_contents ("/usr/share/php/domframework/verify.php"),
      "class")));
eval (str_replace (array ("getopts", "\\verify"),
                   array ("DPT_getopts", "\\DPT_verify"),
      strstr (file_get_contents ("/usr/share/php/domframework/getopts.php"),
      "class")));

/** Manage the getopt options (overwrite the configuration if exists)*/
$getopts = new DPT_getopts ();
$getopts->add ("Help", "?h", "help", "Help on the domphptests soft");
$getopts->add ("Bootstrap", "", "bootstrap:", "Add a bootstrap file to load",
                                "bootstrapFile");
$getopts->add ("Color", "", "color", "Display the results in color");
$getopts->add ("Debug", "d", "debug", "Debug the result");
$getopts->add ("NoFork", "", "nofork", "Do not fork and parallelize the tests");
$getopts->add ("Stats", "", "statistics", "Display the statistics");
$getopts->add ("Time", "", "time", "Display the time need to test");
$getopts->add ("NoIni", "", "noini",
               "Do not use the domphptests.ini file if exists ");
$getopts->add ("CoverStats", "", "coverStats", "Display the cover stats");
$getopts->add ("CoverPath", "", "coverPath:",
               "The path to store the HTML coverage reports", "CoverPath");

/** Default Configuration */
$fork = true;
$debug = false;
$bootstrap = null;
$time = false;
$color = false;
$fileTests = array ();
$statistics = false;
$coverStats = false;
$coverPath = false;
/** Read the configuration file if exists */
if ($getopts->get ("NoIni") === false && file_exists ("domphptests.ini"))
{
  $inivars = parse_ini_file ("domphptests.ini");
  if (isset ($inivars["testfiles"]) && is_array ($inivars["testfiles"]))
  {
    foreach ($inivars["testfiles"] as $test)
    {
      $fileTests = array_merge ($fileTests, glob ($test));
    }
  }
  if (isset ($inivars["bootstrap"]))
    $bootstrap = $inivars["bootstrap"];
  if (isset ($inivars["color"]))
    $color = !! $inivars["color"];
  if (isset ($inivars["debug"]))
    $debug = !! $inivars["debug"];
  if (isset ($inivars["nofork"]))
    $nofork = !! $inivars["nofork"];
  if (isset ($inivars["time"]))
    $time = !! $inivars["time"];
  if (isset ($inivars["statistics"]))
    $statistics = !! $inivars["statistics"];
  if (isset ($inivars["coverStats"]))
    $coverStats = !! $inivars["coverStats"];
  if (isset ($inivars["coverPath"]))
    $coverPath = $inivars["coverPath"];
}

if ($getopts->get ("Help"))
  die ($getopts->help ());
if ($getopts->get ("Bootstrap"))
  $bootstrap = $getopts->get ("Bootstrap");
if ($getopts->get ("Color"))
  $color = true;
if ($getopts->get ("Debug"))
  $debug = true;
if ($getopts->get ("NoFork"))
  $fork = false;
if ($getopts->get ("Stats"))
  $statistics = true;
if ($getopts->get ("Time"))
  $time = true;
if ($getopts->get ("CoverStats"))
  $coverStats = true;
if ($getopts->get ("CoverPath"))
  $coverPath = $getopts->get ("CoverPath");
if (count ($getopts->restOfLine ()))
  $fileTests = $getopts->restOfLine ();
if ($bootstrap !== null)
{
  if (! file_exists ($bootstrap))
    die ("File autoload '$bootstrap' not found\n");
  if (! is_readable ($bootstrap))
    die ("File autoload '$bootstrap' not readable\n");
  if ($debug) echo "Loading $bootstrap file\n";
  require_once ($bootstrap);
}
if ($coverPath && ! function_exists ("xdebug_start_code_coverage"))
  die ("Can not generate the coverage report if xdebug is not installed\n");
if ($coverPath && ! file_exists ($coverPath))
  die ("Not existing path '$coverPath'\n");
$baseDir = realpath (getcwd ());

$fileTests = array_unique ($fileTests);
if (count ($fileTests) > 0)
{
  foreach ($fileTests as $fileTest)
  {
    if (! file_exists ($fileTest))
    {
      file_put_contents ("php://stderr", "File '$fileTest' not found\n");
      exit;
    }
    if (! is_readable ($fileTest))
    {
      file_put_contents ("php://stderr", "File '$fileTest' not readable\n");
      exit;
    }
  }

  $errorlogFiles = array ();
  $statFiles = array ();
  if (function_exists ("pcntl_fork") && $fork === true)
  {
    $childs = array ();
    $results = array ();
    foreach ($fileTests as $fileTest)
    {
      // Fork each test file in a child process to parrallelize to work
      // Define the value to be suire to see it if there is a crash
      $errorlogFile = tempnam (sys_get_temp_dir (), "errorlogfile");
      $statFile = tempnam (sys_get_temp_dir (), "statFile");
      $errorlogFiles[$fileTest] = $errorlogFile;
      $statFiles[$fileTest] = $statFile;
      $results[$fileTest] = 255;
      $pid = pcntl_fork();
      if ($pid == -1)
        die('could not fork');
      elseif ($pid)
        $childs[$fileTest] = $pid;
      else
      {
        //echo "File '$fileTest' forked\n";
        $DomPHPTests = new DomPHPTests ();
        $DomPHPTests->debug = $debug;
        $DomPHPTests->time = $time;
        $DomPHPTests->color = $color;
        $DomPHPTests->coverStats = $coverStats;
        $DomPHPTests->coverPath = $coverPath;
        $DomPHPTests->baseDir = $baseDir;
        $DomPHPTests->bootstrap = $bootstrap;
        exit ($DomPHPTests->main ($fileTest, $errorlogFile, $statFile));
        //echo "End of file $fileTest\n";
        exit;
      }
    }
    while (count($childs) > 0)
    {
      foreach ($childs as $key => $pid)
      {
        $res = pcntl_waitpid ($pid, $status, WNOHANG);
        if($res == -1 || $res > 0)
        {
          $results[$key] = pcntl_wexitstatus ($status);
          unset($childs[$key]);
        }
      }
      usleep(100000);
    }
  }
  else
  {
    // No fork support : sequential support of the tests
    foreach ($fileTests as $fileTest)
    {
      //echo "File '$fileTest' sequenced\n";
      $errorlogFile = tempnam (sys_get_temp_dir (), "errorlogfile");
      $statFile = tempnam (sys_get_temp_dir (), "statFile");
      $errorlogFiles[$fileTest] = $errorlogFile;
      $statFiles[$fileTest] = $statFile;
      $DomPHPTests = new DomPHPTests ();
      $DomPHPTests->debug = $debug;
      $DomPHPTests->time = $time;
      $DomPHPTests->color = $color;
      $DomPHPTests->coverStats = $coverStats;
      $DomPHPTests->coverPath = $coverPath;
      $DomPHPTests->baseDir = $baseDir;
      $DomPHPTests->bootstrap = $bootstrap;
      $results[$fileTest] = $DomPHPTests->main ($fileTest, $errorlogFile,
                                                $statFile);
    }
  }

  ////////////////////////////////////////////
  //   END OF TESTS : DISPLAY THE RESULTS   //
  ////////////////////////////////////////////
  //ksort ($results);
  //var_dump ($results);
  $Nberrors = 0;
  $crashes = array ();
  foreach ($results as $key=>$val)
  {
    if ($val === 255)
     $crashes[$key] = $val;
    else
      $Nberrors += $val;
  }
  $statMemory = array ();
  $statExecuteTime = array ();
  $statTestsNb = array ();
  $statCoverage = array ();
  $cover = array ();
  $coverTests = array ();
  $TestsNbTotal = 0;
  foreach ($statFiles as $fileTmp=>$statFile)
  {
    $stat = json_decode (file_get_contents ($statFile), true);
    $statMemory[$fileTmp] = $stat["MaxMemory"];
    $statExecuteTime[$fileTmp] = $stat["ExecuteTime"];
    $statTestsNb[$fileTmp] = $stat["TestsNb"];
    $TestsNbTotal += $stat["TestsNb"];
    $statCoverage[$fileTmp] = $stat["Coverage"];
    if (isset ($stat["Coverage"]["Cover"]))
    {
      foreach ($stat["Coverage"]["Cover"] as $file=>$values)
      {
        if (! key_exists ($file, $cover))
          $cover[$file] = $values;
        // Merge the local coverage to the main cover. If the old values was
        // -1 (not tested), and the new is 1 (tested), update the $cover
        foreach ($values as $line=>$cov)
        {
          if ($cov === 1 && (! key_exists ($file, $cover) ||
              ! key_exists ($line, $cover[$file]) ||
              $cover[$file][$line] !== 1))
          {
            $cover[$file][$line] = $cov;
          }
        }
      }
    }
    if (isset ($stat["Coverage"]["CoverTests"]))
    {
      foreach ($stat["Coverage"]["CoverTests"] as $file=>$values)
      {
        if (! key_exists ($file, $coverTests))
          $coverTests[$file] = array ();
        foreach ($values as $line=>$fileTest)
        {
          $coverTests[$file][$line][] = $fileTest[0];
        }
      }
    }
    unlink ($statFile);
  }
  echo "\n";
  foreach ($errorlogFiles as $file=>$errorlogFile)
  {
    echo file_get_contents ($errorlogFile);
    unlink ($errorlogFile);
  }
  if ($color)
  {
    if ($Nberrors > 0 || count ($crashes) > 0)
      echo "\033[1;37m\033[41m";
    else
      echo "\033[1;37m\033[42m";
  }
  echo "\nEnd of all tests : $Nberrors errors on $TestsNbTotal tests";
  if (count ($crashes))
  {
    echo " and ".count ($crashes)." crashes\n";
  }
  else
    echo "\n";
  if ($color) echo "\033[0m\n";
  foreach ($results as $file=>$error)
  {
    if ($error == 255)
      echo "See $file : test crashed\n";
    elseif ($error > 0)
      echo "See $file : $error errors\n";
  }
  // Display statistics
  if ($statistics)
  {
    echo "Statistics:\n";
    $maxmem = max ($statMemory);
    echo "  Max memory used : ".intval ($maxmem/1024)."KB in ".
            implode (",", array_keys ($statMemory, $maxmem))."\n";
    if ($debug) print_r ($statMemory);
    $maxtime = max ($statExecuteTime);
    echo "  Max Time used : ".round ($maxtime, 3)."s in ".
            implode (",", array_keys ($statExecuteTime, $maxtime))."\n";
    if ($coverStats)
      echo "  Coverage : \n";
    // 1: this line was executed
    // -1: this line was not executed
    // -2: this line did not have executable code on it
    foreach ($cover as $file=>$used)
    {
      $filename = substr ($file, strlen ($baseDir) + 1);
      if (! file_exists ($filename))
        continue;
      $count = array_count_values ($cover[$file]);
      $count["1"] = (key_exists ("1", $count)) ? $count["1"] : 0;
      $count["-1"] = (key_exists ("-1", $count)) ? $count["-1"] : 0;
      $count["-2"] = (key_exists ("-2", $count)) ? $count["-2"] : 0;
      $fileNames = array ();
      $coverage = array (
        "Cover" => $cover,
        "FileNames" => $fileNames,
        "LinesChecked" => $count["1"],
        "LinesNotCheckable" => $count["-2"],
        "LinesNotChecked" => $count["-1"],
        "LinesTotal" => count ($cover[$file]),
        "PercentChecked" =>
          round ((100 * $count["1"]) / count ($cover[$file]), 2),
        "PercentNotCheckable" =>
          round ((100 * $count["-2"]) / count ($cover[$file]), 2),
        "PercentNotChecked" =>
          round ((100 * $count["-1"]) / count ($cover[$file]), 2),
      );
      if ($coverPath)
      {
        $d = "";
        $d .= "<html>\n";
        $d .= "  <head>\n";
        $d .= "    <style>\n";
        $d .= "      body,html {\n";
        $d .= "        font-family: sans-serif;\n";
        $d .= "      }\n";
        $d .= "      .lineNumber {\n";
        $d .= "        display: inline-block;\n";
        $d .= "        width: 3%;\n";
        $d .= "        background-color: #ccf;\n";
        $d .= "      }\n";
        $d .= "      .notChecked {\n";
        $d .= "        background-color: #faa;\n";
        $d .= "      }\n";
        $d .= "      .notCheckable {\n";
        //$d .= "        background-color: #eee;\n";
        $d .= "      }\n";
        $d .= "      .checked {\n";
        $d .= "        background-color: #efe;\n";
        $d .= "      }\n";
        $d .= "      .lineContent {\n";
        $d .= "        display: inline-block;\n";
        $d .= "        width: 80%;\n";
        $d .= "      }\n";
        $d .= "      .testFiles {\n";
        $d .= "        display: inline-block;\n";
        $d .= "        width: 17%;\n";
        $d .= "      }\n";
        $d .= "    </style>\n";
        $d .= "  </head>\n";
        $d .= "  <body>\n";
        $d .= "    <h1>$filename</h1>\n";
        $d .= "<p>Coverage: ".
              ($coverage["PercentChecked"] + $coverage["PercentNotCheckable"]).
              "% (".
              ($coverage["LinesChecked"] + $coverage["LinesNotCheckable"]).
              "/".count ($cover[$file])." lines)".
              "</p>\n";
        $d .= "    <pre>";
        $realFilename = realpath ($filename);
        foreach (file ($filename) as $line=>$content)
        {
          $d .= "<div class='lineNumber'>".($line+1)."</div>";
          if ($cover[$file][$line+1] === -1)
            $d .= "<div class='lineContent notChecked'>";
          elseif ($cover[$file][$line+1] === -2)
            $d .= "<div class='lineContent notCheckable'>";
          else
            $d .= "<div class='lineContent checked'>";
          $d .= rtrim (htmlentities ($content, ENT_QUOTES, "utf-8"));
          $d .= "</div>";
          $d .= "<div class='testFiles'>";
          if (key_exists (($line+1), $coverTests[$realFilename]))
            $d .= implode (",", $coverTests[$realFilename][$line+1]);
          $d .= "</div>";
          $d .= "\n";
        }
        $d .= "    </pre>\n";
        $d .= "  </body>\n";
        $d .= "<html>\n";
        if (strpos ($filename, "/") !== false &&
            ! file_exists ($coverPath."/".dirname ($filename)))
        {
          mkdir ($coverPath."/".dirname ($filename), 0755, true);
        }
        file_put_contents ($coverPath."/".$filename.".html", $d);
      }

      if ($coverStats)
      {
        echo sprintf ("%-70s", $file)." => ";
        echo sprintf ("%' 3d",
             ($coverage["PercentChecked"] + $coverage["PercentNotCheckable"])).
             "%\n";
      }
    }
    if ($debug) print_r ($statExecuteTime);
  }
  exit ($Nberrors + count ($crashes));
}
else
{
  // File is the test itself
  file_put_contents ("php://stderr", "No test file provided\n");
}
